import random

METADATA_EXCHANGE = 1 << 20

PEER_ID = b"-MD100A-" + bytes([random.randint(0, 255) for _ in range(12)])

MAX_PACKET_SIZE = 2 ** 15

EXTENDED_ID_METADATA = 1

DEFAULT_TRACKERS = [
    "udp://tracker.coppersurfer.tk:6969/announce",
    "udp://tracker.leechers-paradise.org:6969/announce",
    "udp://tracker.opentrackr.org:1337/announce",
    "udp://p4p.arenabg.com:1337/announce",
    "udp://9.rarbg.to:2710/announce",
    "udp://9.rarbg.me:2710/announce",
    "udp://tracker.pomf.se:80/announce",
    "udp://tracker.openbittorrent.com:80/announce",
    "udp://exodus.desync.com:6969/announce",
    "udp://tracker.tiny-vps.com:6969/announce",
    "udp://tracker.moeking.me:6969/announce",
    "udp://retracker.lanta-net.ru:2710/announce",
    "udp://open.stealth.si:80/announce",
    "udp://denis.stalker.upeer.me:6969/announce",
    "udp://tracker.torrent.eu.org:451/announce",
    "udp://tracker.cyberia.is:6969/announce",
    "udp://open.demonii.si:1337/announce",
    "udp://ipv4.tracker.harry.lu:80/announce",
    "udp://tracker3.itzmx.com:6961/announce",
    "udp://zephir.monocul.us:6969/announce",
]

BITTORRENT_PORT = random.randint(10000, 60000)

SERVE_APIKEY = None

DHT_BOOTSTRAP_NODES = [  # TODO: make hostnames and resolve on demand
    ("82.221.103.244", 6881),
    ("67.215.246.10", 6881),
    ("212.129.33.59", 6881),
    ("87.98.162.88", 6881),
    ("174.129.43.152", 6881),
]
DHT_PORT = 6881
DHT_SERVER = None
TORRENT_CACHE_FOLDER = None



import asyncio
import struct
from ipaddress import IPv4Address
from urllib.parse import quote

import aiohttp
from yarl import URL

from . import settings
from .bencode import bdecode, BTFailure


async def retrieve_peers_http_tracker(task_registry, tracker, infohash):
    url = f"{tracker}?info_hash={quote(infohash)}&peer_id={quote(settings.PEER_ID)}&port={settings.BITTORRENT_PORT}&uploaded=0&downloaded=0&left=16384&compact=1&event=started&no_peer_id=1&numwant=200"
    failed = False
    async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=7)) as session:
        try:
            async with session.get(URL(url, encoded=True)) as response:
                task = asyncio.ensure_future(response.read())
                task_registry.add(task)
                result = await task
                if response.status != 200:
                    failed = True
                task_registry.remove(task)
        except (
            aiohttp.client_exceptions.ClientConnectorError,
            asyncio.TimeoutError,
            asyncio.CancelledError,
        ):
            failed = True

    if failed:
        return tracker, {"seeders": 0, "leechers": 0, "peers": []}

    try:
        result = bdecode(result)
    except BTFailure:
        failed = True

    if failed or b"failure reason" in result:
        return tracker, {"seeders": 0, "leechers": 0, "peers": []}

    peer_data = result[b"peers"]
    peers = []
    while len(peer_data) >= 6:
        peer_ip, peer_port = struct.unpack("!IH", peer_data[:6])
        peers.append((IPv4Address(peer_ip), peer_port))
        peer_data = peer_data[6:]

    return (
        tracker,
        {
            "seeders": result[b"complete"],
            "leechers": result[b"incomplete"],
            "peers": peers,
        },
    )



import random

from expiringdict import ExpiringDict


class ForgetfulPeerStorage:
    def __init__(self, ttl=3600):
        self._ttl = ttl
        self._data = ExpiringDict(max_age_seconds=ttl, max_len=2000)

    def get_peers(self, info_hash):
        if info_hash not in self._data:
            return []
        return self._data[info_hash].keys()

    def insert_peer(self, info_hash, peer):
        self.data[info_hash][peer] = None
        self.data[info_hash] = self.data[info_hash]


class ForgetfulTokenStorage:
    def __init__(self, ttl=600):
        self._data = ExpiringDict(max_age_seconds=ttl, max_len=2000)

    def get_token(self, sender, id, info_hash):
        token = bytes([random.randint(0, 255) for _ in range(16)])
        self._data[token] = (sender[0], id, info_hash)
        return token

    def verify_token(self, sender, id, info_hash, token):
        if self._data.get(token) == (sender[0], id, info_hash):
            del self._data[token]
            return True
        return False



# The contents of this file are subject to the BitTorrent Open Source License
# Version 1.1 (the License).  You may not copy or use this file, in either
# source code or executable form, except in compliance with the License.  You
# may obtain a copy of the License at http://www.bittorrent.com/license/.
#
# Software distributed under the License is distributed on an AS IS basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
# for the specific language governing rights and limitations under the
# License.

# Written by Petru Paler
# Modified to have Python 3 support by Anders Jensen

class BTFailure(Exception):
    pass


def decode_int(x, f):
    f += 1
    newf = x.find(b"e", f)
    n = int(x[f:newf])
    if x[f] == 45:
        if x[f + 1] == 48:
            raise ValueError
    elif x[f] == 48 and newf != f + 1:
        raise ValueError
    return (n, newf + 1)


def decode_string(x, f):
    colon = x.find(b":", f)
    n = int(x[f:colon])
    if x[f] == 48 and colon != f + 1:
        raise ValueError
    colon += 1
    return (x[colon : colon + n], colon + n)


def decode_list(x, f):
    r, f = [], f + 1
    while x[f] != 101:
        v, f = decode_func[x[f]](x, f)
        r.append(v)
    return (r, f + 1)


def decode_dict(x, f):
    r, f = {}, f + 1
    while x[f] != 101:
        k, f = decode_string(x, f)
        r[k], f = decode_func[x[f]](x, f)
    return (r, f + 1)


decode_func = {}
decode_func[108] = decode_list
decode_func[100] = decode_dict
decode_func[105] = decode_int

for i in range(48, 59):
    decode_func[i] = decode_string


def bdecode(x):
    try:
        r, l = decode_func[x[0]](x, 0)
    except (IndexError, KeyError, ValueError):
        raise BTFailure("not a valid bencoded string")
    if l != len(x):
        raise BTFailure("invalid bencoded value (data after valid prefix)")
    return r


class Bencached(object):
    __slots__ = ["bencoded"]

    def __init__(self, s):
        self.bencoded = s


def encode_bencached(x, r):
    r.append(x.bencoded)


def encode_int(x, r):
    r.extend((b"i", str(x).encode(), b"e"))


def encode_bool(x, r):
    if x:
        encode_int(1, r)
    else:
        encode_int(0, r)


def encode_string(x, r):
    r.extend((str(len(x)).encode(), b":", x))


def encode_list(x, r):
    r.append(b"l")
    for i in x:
        encode_func[type(i)](i, r)
    r.append(b"e")


def encode_dict(x, r):
    r.append(b"d")
    for k, v in sorted(x.items()):
        r.extend((str(len(k)).encode(), b":", k))
        encode_func[type(v)](v, r)
    r.append(b"e")


encode_func = {}
encode_func[Bencached] = encode_bencached
encode_func[int] = encode_int
encode_func[str] = encode_string
encode_func[bytes] = encode_string
encode_func[list] = encode_list
encode_func[tuple] = encode_list
encode_func[dict] = encode_dict

try:
    from types import BooleanType

    encode_func[BooleanType] = encode_bool
except ImportError:
    pass


def bencode(x):
    r = []
    encode_func[type(x)](x, r)
    return b"".join(r)



import asyncio
import heapq
import logging
import operator
import time
from collections import OrderedDict
from itertools import chain

from .utils import bytes_to_bit_string, shared_prefix

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class FailedToFindBucketException(Exception):
    """Failed to find a bucket for a given node-id, should not happen"""


class KBucket:
    def __init__(self, rangeLower, rangeUpper, ksize):
        self.range = (rangeLower, rangeUpper)
        self.nodes = OrderedDict()
        self.replacement_nodes = OrderedDict()
        self.touch_last_updated()
        self.ksize = ksize

    def touch_last_updated(self):
        self.last_updated = time.monotonic()

    def get_nodes(self):
        return list(self.nodes.values())

    def split(self):
        midpoint = (self.range[0] + self.range[1]) // 2
        one = KBucket(self.range[0], midpoint, self.ksize)
        two = KBucket(midpoint + 1, self.range[1], self.ksize)
        nodes = chain(self.nodes.values(), self.replacement_nodes.values())
        for node in nodes:
            bucket = one if node.long_id <= midpoint else two
            bucket.add_node(node)

        return (one, two)

    def remove_node(self, node):
        if node.id in self.replacement_nodes:
            del self.replacement_nodes[node.id]

        if node.id in self.nodes:
            del self.nodes[node.id]

            if self.replacement_nodes:
                newnode_id, newnode = self.replacement_nodes.popitem()
                self.nodes[newnode_id] = newnode

    def has_in_range(self, node):
        return self.range[0] <= node.long_id <= self.range[1]

    def is_new_node(self, node):
        return node.id not in self.nodes

    def add_node(self, node):
        """
        Add a C{Node} to the C{KBucket}.  Return True if successful,
        False if the bucket is full.

        If the bucket is full, keep track of node in a replacement list,
        per section 4.1 of the paper.
        """
        if node.id in self.nodes:
            del self.nodes[node.id]
            self.nodes[node.id] = node
        elif len(self) < self.ksize:
            self.nodes[node.id] = node
        else:
            if node.id in self.replacement_nodes:
                del self.replacement_nodes[node.id]
            self.replacement_nodes[node.id] = node
            while len(self.replacement_nodes) > self.ksize * 2:
                self.replacement_nodes.popitem(last=False)
            return False
        return True

    def depth(self):
        vals = self.nodes.values()
        sprefix = shared_prefix([bytes_to_bit_string(n.id) for n in vals])
        return len(sprefix)

    def head(self):
        return list(self.nodes.values())[0]

    def __getitem__(self, node_id):
        return self.nodes.get(node_id, None)

    def __len__(self):
        return len(self.nodes)


class TableTraverser:
    def __init__(self, table, startNode):
        index = table.get_bucket_for(startNode)
        table.buckets[index].touch_last_updated()
        self.current_nodes = table.buckets[index].get_nodes()
        self.left_buckets = table.buckets[:index]
        self.right_buckets = table.buckets[(index + 1) :]
        self.left = True

    def __iter__(self):
        return self

    def __next__(self):
        """
        Pop an item from the left subtree, then right, then left, etc.
        """
        if self.current_nodes:
            return self.current_nodes.pop()

        if self.left and self.left_buckets:
            self.current_nodes = self.left_buckets.pop().get_nodes()
            self.left = False
            return next(self)

        if self.right_buckets:
            self.current_nodes = self.right_buckets.pop(0).get_nodes()
            self.left = True
            return next(self)

        raise StopIteration


class RoutingTable:
    def __init__(self, protocol, ksize, node, buckets=None):
        """
        @param node: The node that represents this server.  It won't
        be added to the routing table, but will be needed later to
        determine which buckets to split or not.
        """
        self.node = node
        self.protocol = protocol
        self.ksize = ksize
        if buckets:
            self.buckets = buckets
        else:
            self.flush()

    def flush(self):
        self.buckets = [KBucket(0, 2 ** 160, self.ksize)]

    def split_bucket(self, index):
        one, two = self.buckets[index].split()
        self.buckets[index] = one
        self.buckets.insert(index + 1, two)

    def lonely_buckets(self):
        """
        Get all of the buckets that haven't been updated in over
        an hour.
        """
        hrago = time.monotonic() - 880
        return [b for b in self.buckets if b.last_updated < hrago]

    def remove_contact(self, node):
        index = self.get_bucket_for(node)
        self.buckets[index].remove_node(node)

    def is_new_node(self, node):
        index = self.get_bucket_for(node)
        return self.buckets[index].is_new_node(node)

    def add_contact(self, node):
        index = self.get_bucket_for(node)
        bucket = self.buckets[index]

        # this will succeed unless the bucket is full
        if bucket.add_node(node):
            return

        # Per section 4.2 of paper, split if the bucket has the node
        # in its range or if the depth is not congruent to 0 mod 5
        if bucket.has_in_range(self.node):  # or bucket.depth() % 5 != 0:
            self.split_bucket(index)
            self.add_contact(node)
        else:
            asyncio.ensure_future(self.protocol.call_ping(bucket.head()))

    def get_bucket_for(self, node):
        """
        Get the index of the bucket that the given node would fall into.
        """
        for index, bucket in enumerate(self.buckets):
            if node.long_id <= bucket.range[1]:
                return index
        raise FailedToFindBucketException(f"Failed to find bucket for {node.long_id!r}")

    def find_neighbors(self, node, k=None, exclude=None):
        k = k or self.ksize
        nodes = []
        for neighbor in TableTraverser(self, node):
            notexcluded = exclude is None or not neighbor.same_home_as(exclude)
            if neighbor.id != node.id and notexcluded:
                heapq.heappush(nodes, (node.distance_to(neighbor), neighbor))
            if len(nodes) == k:
                break

        return list(map(operator.itemgetter(1), heapq.nsmallest(k, nodes)))



import asyncio
import logging
import random
import socket
import struct
from ipaddress import IPv4Address

from . import settings

logger = logging.getLogger(__name__)


class TrackerUDPProtocol(asyncio.DatagramProtocol):
    def __init__(self, cb, infohash):
        self.state = "connect"
        self.transport = None
        self.last_transaction_id = None
        self.cb = cb
        self.infohash = infohash

    def get_transaction_id(self):
        self.last_transaction_id = random.randint(0, 2 ** 32 - 1)
        return self.last_transaction_id

    def connection_made(self, transport):
        self.transport = transport
        self.send_connect()

    def send_connect(self):
        data = struct.pack("!qiI", 0x41727101980, 0, self.get_transaction_id())
        self.transport.sendto(data)

    def send_announce(self):
        data = struct.pack(
            "!qiI20s20sqqqiiiiH",
            self.connection_id,
            1,
            self.get_transaction_id(),
            self.infohash,
            settings.PEER_ID,
            0,
            0,
            0,
            0,
            0,
            0,  # Key, not sure
            100,
            0,
        )
        self.transport.sendto(data)

    def datagram_received(self, data, addr):
        logger.debug(f"received datagram from {addr}")
        asyncio.ensure_future(self._handle_response(data, addr))

    async def _handle_response(self, data, addr):
        if self.state == "connect":
            fmt_header = "!iIq"
            if len(data) != struct.calcsize(fmt_header):
                logger.warning("Wrong stuff returned on connect")
                return
            action, transaction_id, connection_id = struct.unpack(fmt_header, data)
            self.connection_id = connection_id
            self.state = "announce"
            self.send_announce()
        elif self.state == "announce":
            fmt_header = "!iIiii"
            header_size = struct.calcsize(fmt_header)
            if len(data) < header_size:
                logger.warning("Wrong stuff returned on announce")
                return
            action, transaction_id, interval, leechers, seeders = struct.unpack(
                fmt_header, data[:header_size]
            )
            peer_data = data[header_size:]
            peers = []
            fmt_peer = "!IH"
            peer_size = struct.calcsize(fmt_peer)
            while len(peer_data) >= peer_size:
                peer_ip, peer_port = struct.unpack(fmt_peer, peer_data[:peer_size])
                peers.append((IPv4Address(peer_ip), peer_port))
                peer_data = peer_data[peer_size:]

            if not self.cb.done():
                self.cb.set_result(
                    {"seeders": seeders, "leechers": leechers, "peers": peers}
                )


async def retrieve_peers_udp_tracker(task_registry, host, port, tracker, infohash):
    loop = asyncio.get_running_loop()
    cb = loop.create_future()
    try:
        transport, protocol = await loop.create_datagram_endpoint(
            lambda: TrackerUDPProtocol(cb, infohash), remote_addr=(host, port)
        )
    except (socket.gaierror, OSError):
        return tracker, {"seeders": 0, "leechers": 0, "peers": []}
    try:
        task = asyncio.ensure_future(cb)
        task_registry.add(task)
        result = await asyncio.wait_for(task, timeout=12)
        task_registry.remove(task)
    except asyncio.TimeoutError:
        return tracker, {"seeders": 0, "leechers": 0, "peers": []}
    except asyncio.CancelledError:
        transport.close()
    else:
        return tracker, result



import asyncio
import logging
import struct
from ipaddress import IPv4Address

from .node import Node, NodeHeap

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


# pylint: disable=too-few-public-methods
class SpiderCrawl:
    """
    Crawl the network and look for given 160-bit keys.
    """

    def __init__(self, protocol, node, peers, ksize, alpha):
        """
        Create a new C{SpiderCrawl}er.

        Args:
            protocol: A :class:`~magnet2torrent.dht.protocol.KRPCProtocol` instance.
            node: A :class:`~magnet2torrent.dht.node.Node` representing the key we're
                  looking for
            peers: A list of :class:`~magnet2torrent.dht.node.Node` instances that
                   provide the entry point for the network
            ksize: The value for k based on the paper
            alpha: The value for alpha based on the paper
        """
        self.protocol = protocol
        self.ksize = ksize
        self.alpha = alpha
        self.node = node
        self.nearest = NodeHeap(self.node, self.ksize)
        self.last_ids_crawled = []
        self.cancel_crawl = False
        self.crawl_finished = False
        log.info("creating spider with peers: %s", peers)
        self.nearest.push(peers)

    async def _find(self, rpcmethod):
        log.info("crawling network with nearest: %s", str(tuple(self.nearest)))

        tasks = set()
        task_mapping = {}
        while not self.cancel_crawl and (
            not self.nearest.have_contacted_all() or tasks
        ):
            count = self.alpha - len(tasks)
            for peer in self.nearest.get_uncontacted()[:count]:
                self.nearest.mark_contacted(peer)
                task = asyncio.ensure_future(rpcmethod(peer, self.node))
                task_mapping[task] = peer.id
                tasks.add(task)

            done, tasks = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
            done_mapping = {}
            for task in done:
                done_mapping[task_mapping.pop(task)] = task.result()
            await self._nodes_found(done_mapping)

        self.crawl_finished = True

        for task in tasks:
            task.cancel()

        return await self._return_value()

    async def _return_value(self):
        raise NotImplementedError

    async def _nodes_found(self, responses):
        raise NotImplementedError


class PeerSpiderCrawl(SpiderCrawl):
    def __init__(self, protocol, node, peers, ksize, alpha, queue):
        SpiderCrawl.__init__(self, protocol, node, peers, ksize, alpha)
        self._queue = queue

    async def find(self):
        """
        Find either the closest nodes or the value requested.
        """
        return await self._find(self.protocol.call_get_peers)

    async def _nodes_found(self, responses):
        """
        Handle the result of an iteration in _find.
        """
        if self.cancel_crawl:
            return

        toremove = []
        for peerid, response in responses.items():
            response = RPCFindResponse(response)
            if not response.happened():
                toremove.append(peerid)
            elif response.has_value():
                await self._queue.put(response.get_values())
            else:
                self.nearest.push(response.get_node_list())
        self.nearest.remove(toremove)

    async def _return_value(self):
        await self._queue.put([])
        return


class NodeSpiderCrawl(SpiderCrawl):
    async def find(self):
        """
        Find the closest nodes.
        """
        return await self._find(self.protocol.call_find_node)

    async def _nodes_found(self, responses):
        """
        Handle the result of an iteration in _find.
        """
        toremove = []
        for peerid, response in responses.items():
            response = RPCFindResponse(response)
            if not response.happened():
                toremove.append(peerid)
            else:
                self.nearest.push(response.get_node_list())
        self.nearest.remove(toremove)

    async def _return_value(self):
        return list(self.nearest)


class RPCFindResponse:
    def __init__(self, response):
        """
        A wrapper for the result of a RPC find.
        """
        self.response = response

    def happened(self):
        """
        Did the other host actually respond?
        """
        return self.response[0]

    def has_value(self):
        return b"values" in self.response[1]

    def get_values(self):
        peers = []
        for value in self.response[1].get(b"values", []):
            peer_ip, peer_port = struct.unpack("!IH", value)
            peers.append((IPv4Address(peer_ip), peer_port))
        return peers

    def get_node_list(self):
        """
        Get the node list in the response.  If there's no value, this should
        be set.
        """
        response = self.response[1].get(b"nodes")
        nodelist = []
        while response and len(response) >= 26:
            peer_id = response[:20]
            peer_ip, peer_port = struct.unpack("!IH", response[20:26])
            node = Node(peer_id, str(IPv4Address(peer_ip)), peer_port)
            nodelist.append(node)
            response = response[26:]
        return nodelist



import base64
from urllib.parse import quote

from aiohttp import web

from . import settings
from .exceptions import FailedToFetchException
from .magnet2torrent import Magnet2Torrent

routes = web.RouteTableDef()


@routes.get("/")
async def get_torrent(request):
    if (
        settings.SERVE_APIKEY
        and request.query.getone("apikey", None) != settings.SERVE_APIKEY
    ):
        raise web.HTTPUnauthorized()

    try:
        magnet = request.query["magnet"]
    except:
        return web.json_response(
            {"status": "error", "message": "magnet argument missing from url"},
            status=400,
        )

    m2t = Magnet2Torrent(
        magnet,
        dht_server=settings.DHT_SERVER,
        torrent_cache_folder=settings.TORRENT_CACHE_FOLDER,
    )
    try:
        filename, torrent_data = await m2t.retrieve_torrent()
    except FailedToFetchException:
        return web.json_response(
            {"status": "error", "message": "failed to retrieve magnet link"}, status=500
        )

    if request.query.getone("direct", None) is not None:
        return web.Response(
            body=torrent_data,
            headers={
                "Content-Disposition": f"attachment; filename*=UTF-8''{quote(filename)}"
            },
        )
    else:
        return web.json_response(
            {
                "status": "success",
                "filename": filename,
                "torrent_data": base64.b64encode(torrent_data).decode("utf-8"),
            }
        )



import asyncio
import hashlib
import logging
import os
import random
import struct
from base64 import b64encode
from ipaddress import IPv4Address

from ..bencode import BTFailure, bdecode, bencode
from .node import Node
from .routing import RoutingTable

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Some code taken from
# https://github.com/bmuller/rpcudp/blob/master/rpcudp/protocol.py
# Has same license as kademlia, see __init__.py

MIN_ID = 0
MAX_ID = 2 ** 160


class KRPCProtocol(asyncio.DatagramProtocol):
    def __init__(
        self,
        source_node,
        peer_storage,
        token_storage,
        ksize,
        wait_timeout=5,
        buckets=None,
    ):
        self.router = RoutingTable(self, ksize, source_node, buckets=buckets)
        self.peer_storage = peer_storage
        self.token_storage = token_storage
        self.source_node = source_node
        self._wait_timeout = wait_timeout
        self._outstanding = {}
        self.transport = None

    def connection_made(self, transport):
        self.transport = transport

    def datagram_received(self, data, addr):
        log.debug("received datagram from %s", addr)
        try:
            data = bdecode(data)
        except (BTFailure, ValueError, KeyError):
            log.info("Failed to decode message")
            return

        if not isinstance(data, dict):
            return

        query_type = data.get(b"y")
        transaction_id = data.get(b"t")
        if not transaction_id:
            return

        if query_type == b"q":  # query
            args = data.get(b"a")
            func_name = data.get(b"q")
            if func_name and isinstance(args, dict):
                asyncio.ensure_future(
                    self.handle_request(transaction_id, func_name, args, addr)
                )
        elif query_type == b"r":  # response
            args = data.get(b"r")
            if isinstance(args, dict):
                asyncio.ensure_future(self.handle_response(transaction_id, args, addr))
        else:
            return

    async def handle_request(self, transaction_id, func_name, args, addr):
        func = getattr(self, f"rpc_{func_name.decode('utf-8')}", None)
        if func is None or not callable(func):
            msgargs = (self.__class__.__name__, func_name)
            log.info("%s has no callable method " "rpc_%s; ignoring request", *msgargs)
            return

        if not asyncio.iscoroutinefunction(func):
            func = asyncio.coroutine(func)
        args = {k.decode("utf-8"): v for (k, v) in args.items()}
        response = await func(addr, **args)
        if response is not None:
            log.debug(
                "sending response %s for msg id %s to %s",
                response,
                b64encode(transaction_id),
                addr,
            )
            txdata = bencode({b"y": b"r", b"r": response,})
            self.transport.sendto(txdata, addr)

    async def handle_response(self, transaction_id, args, addr):
        msgargs = (b64encode(transaction_id), addr)
        if transaction_id not in self._outstanding:
            log.info("received unknown message %s " "from %s; ignoring", *msgargs)
            return
        log.debug("received response %s for message " "id %s from %s", args, *msgargs)
        future, timeout = self._outstanding[transaction_id]
        timeout.cancel()
        if not future.cancelled():
            future.set_result((True, args))
        del self._outstanding[transaction_id]

    def _timeout(self, transaction_id):
        args = (b64encode(transaction_id), self._wait_timeout)
        log.info("Did not received reply for msg " "id %s within %i seconds", *args)
        future = self._outstanding[transaction_id][0]
        if not future.cancelled():
            future.set_result((False, None))
        del self._outstanding[transaction_id]

    def get_refresh_ids(self):
        """
        Get ids to search for to keep old buckets up to date.
        """
        ids = []
        for bucket in self.router.lonely_buckets():
            rid = random.randint(*bucket.range).to_bytes(20, byteorder="big")
            ids.append(rid)
        return ids

    def is_valid_node_id(self, node):
        return MIN_ID < node.long_id < MAX_ID

    def rpc_ping(self, sender, id):
        source = Node(id, sender[0], sender[1])
        if not self.is_valid_node_id(source):
            return
        self.welcome_if_new(source)
        return {b"id": id}

    def rpc_announce_peer(
        self,
        sender,
        id,
        info_hash,
        port,
        token,
        name=None,
        implied_port=None,
        seed=None,
    ):
        source = Node(id, sender[0], sender[1])
        if not self.is_valid_node_id(source):
            return

        self.welcome_if_new(source)
        if self.token_storage.verify_token(sender[0], id, info_hash, token):
            if implied_port:
                port = sender[1]
            log.debug(
                "got an announce_peer request from %s, storing '%s'",
                sender,
                info_hash.hex(),
            )
            self.peer_storage.insert_peer((sender[0], port))
        else:
            log.debug("Invalid token from %s", sender)
        return {b"id": id}

    def rpc_find_node(self, sender, id, target, want="n4", token=None):
        log.info("finding neighbors of %i in local table", int(id.hex(), 16))
        source = Node(id, sender[0], sender[1])
        if not self.is_valid_node_id(source):
            return

        self.welcome_if_new(source)
        node = Node(target)
        if not self.is_valid_node_id(node):
            return
        neighbors = self.router.find_neighbors(node, exclude=source)
        data = {b"id": id, b"nodes": b"".join([n.packed for n in neighbors])}
        if token:
            data[b"token"] = token
        return data

    def rpc_get_peers(
        self, sender, id, info_hash, want="n4", noseed=0, scrape=0, bs=None
    ):
        source = Node(id, sender[0], sender[1])
        if not self.is_valid_node_id(source):
            return
        self.welcome_if_new(source)
        peers = self.peer_storage.get_peers(info_hash)
        token = self.token_storage.get_token(sender, id, info_hash)
        if not peers:
            return self.rpc_find_node(sender, id, info_hash, token=token)

        return {
            b"id": id,
            b"token": token,
            b"values": [
                IPv4Address(peer[0]).packed + struct.pack("!H", peer[1])
                for peer in peers
            ],
        }

    async def call_find_node(self, node_to_ask, node_to_find):
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.find_node(
            address, {b"id": self.source_node.id, b"target": node_to_find.id}
        )
        return self.handle_call_response(result, node_to_ask)

    async def call_get_peers(self, node_to_ask, node_to_find):
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.get_peers(
            address, {b"id": self.source_node.id, b"info_hash": node_to_find.id}
        )
        return self.handle_call_response(result, node_to_ask)

    async def call_ping(self, node_to_ask):
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.ping(address, {b"id": self.source_node.id})
        return self.handle_call_response(result, node_to_ask)

    async def call_announce_peer(self, node_to_ask, key, value):  # TODO
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.store(address, self.source_node.id, key, value)
        return self.handle_call_response(result, node_to_ask)

    def welcome_if_new(self, node):
        """
        Given a new node, send it all the keys/values it should be storing,
        then add it to the routing table.

        @param node: A new node that just joined (or that we just found out
        about).

        Process:
        For each key in storage, get k closest nodes.  If newnode is closer
        than the furtherst in that list, and the node for this server
        is closer than the closest in that list, then store the key/value
        on the new node (per section 2.5 of the paper)
        """
        if not self.router.is_new_node(node):
            return

        log.info("never seen %s before, adding to router", node)
        self.router.add_contact(node)

    def handle_call_response(self, result, node):
        """
        If we get a response, add the node to the routing table.  If
        we get no response, make sure it's removed from the routing table.
        """
        if not result[0]:
            log.info("no response from %s, removing from router", node)
            self.router.remove_contact(node)
            return result

        log.info("got successful response from %s", node)
        self.welcome_if_new(node)
        return result

    def generate_token(self):
        return bytes([random.randint(0, 255) for _ in range(16)])

    def __getattr__(self, name):
        """
        If name begins with "_" or "rpc_", returns the value of
        the attribute in question as normal.
        Otherwise, returns the value as normal *if* the attribute
        exists, but does *not* raise AttributeError if it doesn't.
        Instead, returns a closure, func, which takes an argument
        "address" and additional arbitrary args (but not kwargs).
        func attempts to call a remote method "rpc_{name}",
        passing those args, on a node reachable at address.
        """
        if name.startswith("_") or name.startswith("rpc_"):
            return getattr(super(), name)

        try:
            return getattr(super(), name)
        except AttributeError:
            pass

        def func(address, args):
            transaction_id = hashlib.sha1(os.urandom(32)).digest()
            txdata = bencode(
                {
                    b"y": b"q",
                    b"t": transaction_id,
                    b"a": args,
                    b"q": name.encode("utf-8"),
                }
            )
            log.debug(
                "calling remote function %s on %s (msgid %s)",
                name,
                address,
                b64encode(transaction_id),
            )
            self.transport.sendto(txdata, address)

            loop = asyncio.get_event_loop()
            if hasattr(loop, "create_future"):
                future = loop.create_future()
            else:
                future = asyncio.Future()
            timeout = loop.call_later(self._wait_timeout, self._timeout, transaction_id)
            self._outstanding[transaction_id] = (future, timeout)
            return future

        return func



import os

from setuptools import find_packages, setup

def readme():
    with open("README.md") as f:
        return f.read()

setup(
    name="magnet2torrent",
    version="1.2.1",
    url="https://github.com/JohnDoee/magnet2torrent",
    author="Anders Jensen",
    author_email="andersandjensen@gmail.com",
    description="Turn a bittorrent magnet links into a .torrent file.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"":"src"},
    include_package_data=True,
    install_requires=["aiohttp==3.*", "expiringdict>=1.2.0"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Internet :: WWW/HTTP",
    ],
    entry_points={
        "console_scripts": ["magnet2torrent = magnet2torrent.__main__:main",]
    },
)



import asyncio
import base64
import binascii
import logging
from itertools import chain
from pathlib import Path
from urllib.parse import parse_qs, urlparse

from . import settings
from .bencode import bdecode, bencode
from .exceptions import FailedToFetchException
from .httptracker import retrieve_peers_http_tracker
from .peer import fetch_from_peer
from .udptracker import retrieve_peers_udp_tracker

logger = logging.getLogger(__name__)


class Magnet2Torrent:
    def __init__(
        self,
        magnet_link,
        use_trackers=True,
        use_additional_trackers=False,
        dht_server=None,
        torrent_cache_folder=None,
    ):
        self.magnet_link = magnet_link
        self.use_trackers = use_trackers
        self.use_additional_trackers = use_additional_trackers
        self.dht_server = dht_server
        self.torrent_cache_folder = torrent_cache_folder

    def _parse_url(self):
        url = urlparse(self.magnet_link)
        url_query = parse_qs(url.query)
        infohash = url_query["xt"][0].split(":")[2]
        if len(infohash) == 40:
            infohash = binascii.unhexlify(infohash)
        elif len(infohash) == 32:
            infohash = base64.b32decode(infohash)
        else:
            raise Exception("Unable to parse infohash")

        trackers = url_query.get("tr", [])
        name = url_query.get("dn")
        if name:
            name = name[0]
        else:
            name = binascii.hexlify(infohash).decode()
        return infohash, trackers, name

    @property
    def infohash(self):
        return self._parse_url()[0]

    @property
    def trackers(self):
        return self._parse_url()[1]

    @property
    def name(self):  # TODO: better stripping
        return (
            self._parse_url()[2]
            .strip(".")
            .replace("/", "")
            .replace("\\", "")
            .replace(":", "")
        )

    def create_torrent(self, torrent_data):
        torrent = {b"info": bdecode(torrent_data)}
        trackers = None
        if self.use_trackers:
            trackers = self.trackers
            if self.use_additional_trackers:
                trackers += settings.DEFAULT_TRACKERS

            torrent[b"announce-list"] = [
                [tracker.encode("utf-8")] for tracker in trackers
            ]
            if torrent[b"announce-list"]:
                torrent[b"announce"] = torrent[b"announce-list"][0][0]

        return f"{self.name}.torrent", bencode(torrent)

    @property
    def torrent_cache_path(self):
        if not self.torrent_cache_folder:
            return None

        filename = binascii.hexlify(self.infohash).decode("utf-8")
        return (
            Path(self.torrent_cache_folder)
            / Path(filename[:2])
            / Path(filename[2:4])
            / Path(filename)
        )

    async def retrieve_torrent(self):
        torrent_cache_path = self.torrent_cache_path
        if torrent_cache_path and torrent_cache_path.exists():
            logger.debug(f"We had a cache at {torrent_cache_path!s}")
            return self.create_torrent(torrent_cache_path.read_bytes())

        task_registry = set()
        infohash = self.infohash
        tasks = []
        if self.use_trackers:
            trackers = self.trackers
            if self.use_additional_trackers:
                trackers += settings.DEFAULT_TRACKERS

            for tracker in trackers:
                logger.debug(f"Trying to fetch peers from {tracker}")
                tracker_url = urlparse(tracker)
                if tracker_url.scheme in ["http", "https"]:
                    task = retrieve_peers_http_tracker(task_registry, tracker, infohash)
                elif tracker_url.scheme in ["udp"]:
                    if ":"  not in tracker_url.netloc:
                        continue
                    host, port = tracker_url.netloc.split(":")
                    task = retrieve_peers_udp_tracker(
                        task_registry, host, port, tracker, infohash
                    )
                else:
                    print(f"Unknown scheme, {tracker_url.scheme}")
                    continue

                task = asyncio.ensure_future(task)
                task.task_type = "tracker"
                tasks.append(task)

        if self.dht_server:
            task = asyncio.ensure_future(
                self.dht_server.find_peers(task_registry, infohash)
            )
            task.task_type = "tracker"
            tasks.append(task)

        handled_peers = set()
        while tasks:
            done, tasks = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                result = task.result()
                if task.task_type == "tracker":
                    for peer in result[1]["peers"]:
                        if peer in handled_peers:
                            continue
                        handled_peers.add(peer)
                        peer_ip, peer_port = peer
                        logger.debug(f"Connecting to {peer_ip}:{peer_port}")
                        peer_task = asyncio.ensure_future(
                            fetch_from_peer(task_registry, peer_ip, peer_port, infohash)
                        )
                        peer_task.task_type = "peer"
                        tasks.add(peer_task)
                    if len(result) > 2:
                        new_task = asyncio.ensure_future(result[2]())
                        new_task.task_type = task.task_type
                        tasks.add(new_task)
                elif task.task_type == "peer":
                    if result:
                        for task in chain(task_registry, tasks):
                            if not task.done():
                                task.cancel()

                        if torrent_cache_path:
                            torrent_cache_path.parent.mkdir(parents=True, exist_ok=True)
                            torrent_cache_path.write_bytes(result)
                        return self.create_torrent(result)

        raise FailedToFetchException()



"""
General catchall for functions that don't make sense as methods.
"""
import asyncio
import hashlib
import operator


async def gather_dict(dic):
    cors = list(dic.values())
    results = await asyncio.gather(*cors)
    return dict(zip(dic.keys(), results))


def digest(string):
    if not isinstance(string, bytes):
        string = str(string).encode("utf8")
    return hashlib.sha1(string).digest()


def shared_prefix(args):
    """
    Find the shared prefix between the strings.

    For instance:

        sharedPrefix(['blahblah', 'blahwhat'])

    returns 'blah'.
    """
    i = 0
    while i < min(map(len, args)):
        if len(set(map(operator.itemgetter(i), args))) != 1:
            break
        i += 1
    return args[0][:i]


def bytes_to_bit_string(bites):
    bits = [bin(bite)[2:].rjust(8, "0") for bite in bites]
    return "".join(bits)
